#!/usr/bin/env python

import numpy as np
import pyMEB
import wind
import pathlib
import scipy.stats
from mpi4py import MPI
        
home = str(pathlib.Path.home())

argv = ['./MEB_online',
        f'--config=EB_channel_adv.cfg',
        '--forcing.forcing-type=External',
        '--input.from-restart=true',
        '--input.restartfile-timeindex=0',
        '--output.dirname=output',
        '--input.dirname=input',
        ]
varnames = ('area', 'cohesion', 'thickness', 'damage', 'u', 'v',
            'stress_xx', 'stress_xy', 'stress_yy', )
global_params = ('alpha', 'C_a',)

varnames_state = ('area', 'cohesion', 'thickness', 'damage', 'u', 'v',
            'stress_xx', 'stress_xy', 'stress_yy', )

global_params_state = ('alpha', )
_default_params_value = {'C_a': 1.5e-3, 'alpha':4.0} # (1.5e-3, 4.0, )
_params_std = {'C_a': 1.e-4, 'alpha': 1.2}
_parms_lowbound = {'C_a': 0., 'alpha': 2.}
_params_filename = {'C_a':'/home/users/ia923171/EXP_MEB/inputs_AC/input_Ca.npy', 'alpha': '/home/users/ia923171/EXP_MEB/inputs_AC/input_alpha.npy',} 
_params_filename = {'C_a':'/home/users/ia923171/EXP_MEB/all/D/input_Ca_tmp.npy', 'alpha': '/home/users/ia923171/EXP_MEB/inputs_AC/input_alpha.npy',} 

doLog = False
doAnomaly = False 

class MEB:
    def __init__(self, para, type_model):
        """initialise the model
        """
        if para.PE[type_model]:
            print (f' initialising {type_model}')
            input_dir_suffix = para.task_id
            argv[-2:] = [f'--output.dirname=output_{type_model}_{para.task_id}',
                         f'--input.dirname=input{input_dir_suffix}']
            self.dg, self.post, self.output, self.dt, self.d = pyMEB.initModel(para.comm_model, argv)
            task_id = 0 if (type_model == 'forecast') else para.task_id
            self.dg.set_taskid(task_id)
            nt = pyMEB.getTimeStep(self.dg)
        else:
            self.dg, self.post, self.output, self.dt, self.d = None, None, None, None, None
            nt = None
        self.nt = para.comm_model.bcast(nt, root=0)

        # it could be forecast, analysis, or ensemble
        self.type_model = type_model
        # model fields 
        self.X = dict()
        # model field/parameters dimensions in state vector
        self.dim = dict()
        self.isModelHead = True if para.rank_model == 0 and para.PE[type_model] else False

        self.initField()
        self.initModelDim()
        self.state = np.empty([self.StateSize], dtype='d') if self.isModelHead else None

        self.mean_params = None

        self.mean_params = dict()
        self.noise = dict()
        if self.dg is not None:
            alpha = np.exp(-self.dt/24./3600/2.)
            beta = np.sqrt(1 - alpha*alpha)
            for varname in global_params:
                self.noise[varname] = 0.
                param = np.load(_params_filename[varname])[task_id-1] + self.noise[varname]
                self.mean_params[varname] = np.mean(np.load(_params_filename[varname]))
                pyMEB.assignValueSerial(self.dg, varname, float(param))
            pyMEB.adjustModelConsistency(self.dg, nt)

        para.comm_world.Barrier()


    def initWind(self, task_id, do_random, do_walk, Lmin=-100000, L=200000, total_steps=90*24*3600):
        """initialise wind field up to the restart file
        """
        if self.dg is None:
            return None

        # initialise storm
        coord = pyMEB.getVarCoordinate(self.dg, 'forcing_x')
        storm = wind.Storm(task_id, *coord, Lmin, L, self.dt, total_steps, do_random, do_walk)
        storm_t = wind.Storm(0, *coord, Lmin, L, self.dt, total_steps, True, True)
        # update to current step
        for it in range(self.nt):
            u, v = storm.updateWind(self.dt, it + 1, self.dt)
            u_t, v_t = storm_t.updateWind(self.dt, it + 1, self.dt)
        # first assign to MEB model        
        norm = np.sqrt(u*u + v*v)
        if 'C_a' in global_params_state:
            Ca = np.load(_params_filename['C_a'])[task_id - 1]
        else:
            Ca = 1.5e-3
            # Ca = np.load('/home/users/ia923171/EXP_MEB/all/D/input_Ca_tmp.npy')[task_id - 1]
        pyMEB.assignValue(self.dg, 'forcing_x', Ca*1.3*norm*u)
        pyMEB.assignValue(self.dg, 'forcing_y', Ca*1.3*norm*v)
        return storm, storm_t


    def initModelDim(self):
        """specify the position of model field in state vector
        """
        if self.dg is None:
            return None

        # get dimension of the state variables
        cumsum = 0
        for varname in varnames_state:
            start = cumsum
            cumsum += pyMEB.getArraySizeTotal(self.dg, varname)
            self.dim[varname] = (start, cumsum)
            if self.isModelHead:
                print(f'...{varname} dim is {self.dim[varname]}...')

        for varname in global_params_state:
            start = cumsum
            cumsum += 1
            self.dim[varname] = (start, cumsum)
            if self.isModelHead:
                print(f'...{varname} dim is {self.dim[varname]}...')


    def initField(self):
        """We construct a model fields in self.X from C++ DG model.
        """
        if self.dg is None:
            return None

        # retrieve initial condition from model state
        for varname in varnames:
            dim = pyMEB.getArraySizeTotal(self.dg, varname)
            self.X[varname] = np.zeros(dim)

        for varname in global_params:
            dim = 1
            self.X[varname] = np.zeros(1)


    def getStateVector(self):
        """get updated model field into state vector"""
        if self.dg is None:
            return None

        f_trans = np.log if doLog else lambda x: x

        if self.isModelHead:
            for varname in varnames_state:
                self.state[range(*self.dim[varname])] = self.X[varname]
            for varname in global_params_state:
                self.state[range(*self.dim[varname])] = f_trans(self.X[varname])

        return self.preprocessing()


    def preprocessing(self):
        """preprocessing model fields to state vector
        """
        return self.state


    def getEnsemble(self, E, N, para):
        """initialise the ensemble
        """
        if self.dg is None:
            return None

        self.state = self.getStateVector()

        # gather state to form ensemble
        if self.state is not None:
            para.comm_couple.Gather(self.state, E, root=0)

        return E


    def assignField(self, state):
        """assign state vector to model fields
        """
        f_trans = (lambda x: np.e**x) if doLog else (lambda x: x)
        for varname in varnames_state:
            self.X[varname] = state[range(*self.dim[varname])]
        for varname in global_params_state:
            self.X[varname] = f_trans(state[range(*self.dim[varname])])


    def assignEnsemble(self, E, it, para):
        """initialise the ensemble
        """
        if self.dg is None:
            return None

        # scatter E to X
        if para.PE['filter'] and doAnomaly:
            for varname in global_params_state:
                E[:, range(*self.dim[varname])] += self.mean_params[varname]
                self.mean_params[varname] = np.mean(E[:, range(*self.dim[varname])])
                E[:, range(*self.dim[varname])] -= self.mean_params[varname]

        self.mean_params = para.comm_couple.bcast(self.mean_params, root=0)
        self.mean_params = para.comm_model.bcast(self.mean_params, root=0)

        if self.isModelHead:
            para.comm_couple.Scatter(E, self.state, root=0)
            self.assignField(self.state)
            self.postprocessing()

        # assign values to DG
        for varname in varnames:
            pyMEB.assignValueSerial(self.dg, varname, self.X[varname])

        for varname in global_params:
            var = self.X[varname][0]
            var = var + self.mean_params[varname] if doAnomaly else var
            pyMEB.assignValueSerial(self.dg, varname, var)
        
        assert isinstance(it, int), 'it must be integer'
        pyMEB.adjustModelConsistency(self.dg, it)


    def postprocessing(self):
        """post-processing model field
        """
        # thickness
        # new_h = np.maximum(0, self.X['area'] - 1.)*self.X['thickness'] + self.X['thickness']
        self.X['thickness'] = np.maximum(self.X['thickness'], 0.)
        # concentration
        self.X['area'] = np.maximum(self.X['area'], 0.)
        self.X['area'] = np.minimum(self.X['area'], 1.)
        # cohesion
        self.X['cohesion'] = np.maximum(self.X['cohesion'], 5000.)
        self.X['cohesion'] = np.minimum(self.X['cohesion'], 10000.)
        # damage
        self.X['damage'] = np.maximum(self.X['damage'], 0.)
        self.X['damage'] = np.minimum(self.X['damage'], 1.)
        if ('C_a' in global_params):
            while True:
                if self.X['C_a'][0] > _parms_lowbound['C_a']: break
                self.X['C_a'][0] = np.random.normal(1e-5, 1e-6)
        if ('alpha' in global_params):
            self.X['alpha'][0] = np.maximum(self.X['alpha'][0], _parms_lowbound['alpha'])


    def updateField(self):
        """update model field from DG object
        """
        if self.dg is None:
            return None

        for varname in varnames:
            self.X[varname] = pyMEB.getValueSerial(self.dg, varname, self.X[varname])
            if np.sum(np.isnan(self.X[varname])) > 0:
                print (f'There are NaNs in the model field {varname} at time step {pyMEB.getTimeStep(self.dg)}')
                MPI.COMM_WORLD.Abort() 

        for varname in global_params:
            self.X[varname][0] = pyMEB.getValueSerial(self.dg, varname, self.X[varname][0])
            if doAnomaly:
                self.X[varname][0] -= self.mean_params[varname]
            if np.sum(np.isnan(self.X[varname])) > 0:
                print (f'There are NaNs in the model field {varname} at time step {pyMEB.getTimeStep(self.dg)}')
                MPI.COMM_WORLD.Abort() 
                

    def updateStateField(self):
        """update model field from DG object
        """
        if self.dg is None:
            return None

        for varname in varnames_state:
            self.X[varname] = pyMEB.getValueSerial(self.dg, varname, self.X[varname])
        for varname in global_params_state:
            self.X[varname][0] = pyMEB.getValueSerial(self.dg, varname, self.X[varname][0])
            if doAnomaly:
                self.X[varname][0] -= self.mean_params[varname]

    def scheme(self, u, u_t, nt, comm_couple, doOutput=True):
        """model forward integration for one ensemble member
        """
        if self.dg is None:
            return None

        for i, uu, uu_t in zip(range(nt), u, u_t):
            ux = uu[0]
            uy = uu[1]
            pyMEB.assignValue(self.dg, 'forcing_x', ux)
            pyMEB.assignValue(self.dg, 'forcing_y', uy)
            pyMEB.stepModel(self.dg, self.post, self.output, self.d)

            if doOutput:
                self.output.update(self.dg, self.post)


    def OutputGen(self, state, it, doPost=False):
        """call output routines
        """
        if self.isModelHead:
            self.assignField(state)
            if doPost:
                self.postprocessing()

        # assign values
        for varname in varnames:
            pyMEB.assignValueSerial(self.dg, varname, self.X[varname])
        for varname in global_params:
            var = self.X[varname][0]
            var = var + self.mean_params[varname] if doAnomaly else var
            pyMEB.assignValueSerial(self.dg, varname, var)

        pyMEB.adjustModelConsistency(self.dg, it)
        self.output.update(self.dg, self.post)


    def OutputGenE(self, E, it, para):
        """make output for ensembles
        """
        if self.dg is None:
            return None

        assert isinstance(it, int), 'it must be integer'

        if para.PE['filter'] and doAnomaly:
            for varname in global_params_state:
                E[:, range(*self.dim[varname])] += self.mean_params[varname]
                self.mean_params[varname] = np.mean(E[:, range(*self.dim[varname])])
                E[:, range(*self.dim[varname])] -= self.mean_params[varname]

        self.mean_params = para.comm_couple.bcast(self.mean_params, root=0)
        self.mean_params = para.comm_model.bcast(self.mean_params, root=0)
        # scatter E to state
        if self.isModelHead:
            para.comm_couple.Scatter(E, self.state, root=0)
        self.OutputGen(self.state, it)

 
    def SaveParams(self, E, t, para, suffix):
        if self.dg is None:
            return None

        if not para.PE['filter'] or not self.isModelHead:
            return None
        
        if len(global_params_state) == 0: return None

        P = dict()
        for varname in global_params_state:
            P[varname] = E[:, range(*self.dim[varname])]
            if doAnomaly:
                P[varname] += self.mean_params[varname]
        np.savez_compressed(f'param_{t}_{suffix}.npz', **P)


    @property
    def StateSize(self):
        """the size of state vector"""
        if self.dg is None:
            return None

        return np.sum([value[1] - value[0] for value in self.dim.values()])


    @property
    def ParamSize(self):
        return len(global_params_state)
