import os

import numpy as np
import scipy.stats

import pyMEB

class Observation:
    def __init__(self, model, gridTypes):
        self.igrid = dict()
        for obsVar, gridType in gridTypes.items():
            self.igrid[obsVar] = self.initObsMesh(gridType, model, obsVar)

        self.dim = dict()
        cumsum = 0
        for varname in model.dim:
            if varname in self.igrid:
                self.dim[varname] = (cumsum, cumsum + np.sum(self.igrid[varname]))
                cumsum += np.sum(self.igrid[varname])


    def initObsMesh(self, gridType, model, varname):
        if model.dg is None:
            return None

        if gridType == 'alongtrack':
            return self.AlongTrack(model, varname)
        elif gridType == 'fulldomain':
            return self.FullDomain(model, varname)
        elif gridType == 'randompatch':
            return self.RandomPatch(model, varname)
        elif gridType == 'sparse':
            return self.SparseObs(model, varname)
        else:
            raise RuntimeError(f'... unimplemented grid type: {gridType}...')

    @classmethod
    def getUstd(cls, yt, condition):
        std = 0.8*np.maximum(np.load('drift_std.npz')['u'], 0.001)
        return std[condition]

    @classmethod
    def getVstd(cls, yt, condition):
        std = 0.8*np.maximum(np.load('drift_std.npz')['v'], 0.001)
        return std[condition]

    @classmethod
    def getSITstd(cls, yt):
        std = np.zeros_like(yt)
        condition = yt < 0.7
        std[condition] = 8
        condition = (yt >= 0.7) & (yt < 3)
        std[condition] = 0.02*yt[condition]*(1 - 7*np.e/(0.3 - yt[condition]**np.e))
        condition = yt >= 3
        std[condition] = 0.01*yt[condition]*(5*(yt[condition] - 3.) + 1. - 7*np.e/(0.3 - 3**np.e))
        return std

    @classmethod
    def getSICstd(cls, yt):
        return np.sqrt(0.01 + (0.5 - np.abs(0.5 - yt))**2)

    @classmethod
    def getSTD(cls, varname, yt, thickness_scale, igrid):
        if varname == 'thickness':
            return cls.getSITstd(yt)*thickness_scale
        elif varname == 'area':
            return cls.getSICstd(yt)
        elif varname == 'u':
            return cls.getUstd(yt, igrid['u'])
        elif varname == 'v':
            return cls.getVstd(yt, igrid['v'])
        else:
            raise RuntimeError(f'... unimplemented observation type: {varname}...')

    @classmethod
    def getObs(cls, varname, yt, std):
        if varname == 'thickness':
            yy = yt + scipy.stats.truncnorm.rvs(-yt, np.inf, loc=0., scale=std)
            assert np.all(yy >= 0.), 'thickness obs is smaller than 0'
        elif varname == 'area':
            yy = yt + scipy.stats.truncnorm.rvs(-yt, 1. - yt, loc=0., scale=std)
            assert np.all(yy >= 0.), 'area obs is smaller than 0'
            assert np.all(yy <= 1.), 'area obs is greater than 1'
        elif varname == 'u' or varname == 'v':
            yy = yt + scipy.stats.truncnorm.rvs(-np.inf, np.inf, loc=0., scale=std)
        else:
            raise RuntimeError(f'... unimplemented observation type: {varname}...')
        return yy

    def genSynthetic(self, para, step, y, std, thickness_scale=1.):
        for varname in self.igrid:
            filename = f'obs/{varname}_{step}.npz'
            f = np.load(filename)
            y[range(*self.dim[varname])] = f['y']
            std[range(*self.dim[varname])] = f['std']
        infl = 1.
        if 'u' in self.igrid:
            std[range(*self.dim['u'])] = infl*std[range(*self.dim['u'])]
            std[range(*self.dim['v'])] = infl*std[range(*self.dim['v'])]
        if 'area' in self.igrid:
            std[range(*self.dim['area'])] = infl*std[range(*self.dim['area'])]
        if 'thickness' in self.igrid:
            std[range(*self.dim['thickness'])] = infl*std[range(*self.dim['thickness'])]
       
        return y, std

    def H(self, E, model):
        if E is None:
            return None

        return np.concatenate([E[..., range(*model.dim[varname])][..., condition]
                   for varname, condition in self.igrid.items()], axis=-1)

    def AlongTrack(self, model, varname):
        x_c, y_c = pyMEB.getVarCoordinate(model.dg, varname)
        condition = np.zeros(len(x_c), dtype=bool)
        for coeff, restriction in zip([2.2, -2.2, 0.5, -0.5], [16000, 16000, 8000, 8000]):
            condition = condition | (np.abs(y_c + coeff*x_c) <= restriction)
        return condition

    def FullDomain(self, model, varname):
        return np.ones(pyMEB.getArraySizeTotal(model.dg, varname), dtype=bool)

    def RandomPatch(self, model, varname):
        condition = self.FullDomain(model, varname)
        M = len(condition)
        condition[np.random.randint(0, M - 1, int(0.4*M))] = False
        return condition

    def SparseObs(self, model, varname):
        return np.load('SparseObs.npy')

    @property
    def ObsDim(self):
        return np.sum([np.sum(condition) for condition in self.igrid.values()])
