#!/usr/bin/env python
# -*- coding: iso-8859-15 -*-h


#################################################################
#
#      (c)   Marc Bocquet, ENPC/CEREA
#      Programme ienks_analysis.py
#      Initial release: 22 June 2015
#      Latest modification: 22 June 2015
#
#      Based on Bocquet and Sakov, 2013-2014 (IEnKS)
#      Based on Bocquet, 2011 and Bocquet & Sakov, 2012 (EnKF-N)
#
#################################################################

from mpi4py import MPI

import numpy as np
cimport numpy as np
import itertools
from numpy import zeros, dot, identity, sqrt, newaxis, diag, eye
# Finite differences scaling factor
cdef double eps = 1.e-4
cdef int p_max = 40
cdef double e = 1.e-3


##############################################################
# SDA, S=1, bundle, IEnKS                                    #
##############################################################
def Analysis1(int t,
              np.ndarray[np.float64_t, ndim=1] y,
              E_in,
              obs,
              tapering,
              int N, int M, models, para, u,
              np.ndarray[np.float64_t, ndim=1] sig_obs, int Lc):

    cdef np.ndarray[np.float64_t, ndim=2] E, Eout, Y, E0
    cdef np.ndarray[np.int_t, ndim=1] chunks
    cdef int chunk, remainder, nPE, rank, i, D, offset

    rank = para.rank_world
    nPE = para.size_world
    chunk = M//nPE
    chunks = np.ones(nPE, dtype=np.int)*chunk
    remainder = M - chunk*nPE
    chunks[:remainder] = chunks[:remainder] + 1
    if rank < remainder:
        chunk += 1
    offset = np.sum(chunks[:rank])


    D = len(sig_obs)
    if para.PE['filter']:
        Eout = np.asfortranarray(E_in)
    else:
        Eout = None

    E = np.empty((N, chunk), dtype=np.float64, order='F')
    E0 = np.empty((N, chunk), dtype=np.float64, order='F')
    Y = np.empty((N, D), dtype=np.float64)
    para.comm_world.Scatterv([Eout, chunks*N], E0, root=para.rank_world_filterpe)

    # Determine size of the ensemble
    cdef np.ndarray[np.float64_t, ndim=2] w0 = zeros((N, chunk))
    cdef np.ndarray[np.float64_t, ndim=1] xm0 = zeros(chunk)
    cdef np.ndarray[np.float64_t, ndim=2] X0 = np.zeros((N, chunk))
    cdef np.ndarray[np.float64_t, ndim=1] _x0 = np.zeros_like(xm0)
    cdef np.ndarray[np.float64_t, ndim=1] x0 = np.zeros_like(xm0)

    xm0 = E0.sum(axis=0)/N
    X0 = E0 - xm0

    cdef int l_domain, p, global_domain
    cdef np.ndarray[np.float64_t, ndim=1] ym, delta, gradient, coeffs, dw0, s
    cdef np.ndarray[np.float64_t, ndim=2] Ysqrt, H, U, V, T
    cdef np.ndarray[np.int_t, ndim=1] l_obs
    cdef np.ndarray[np.float64_t, ndim=1] _e, _e_total
    _e = np.zeros(1, dtype=np.float64) 
    _e_total = np.zeros(1, dtype=np.float64)

    for p in range(p_max):
        for l_domain in range(chunk):
            x0[l_domain] = xm0[l_domain] + dot(w0[:, l_domain], X0[:, l_domain])
            # Compute the sensitivities
            E0[:, l_domain] = x0[l_domain] + eps*X0[:, l_domain]

        # gather E
        para.comm_world.Gatherv(E0, [Eout, chunks*N], root=para.rank_world_filterpe)
        if para.PE['filter']:
            Eout = np.ascontiguousarray(Eout)

        # models['analysis'].assignEnsemble(Eout, t, para)
        # models['analysis'].scheme(u, Lc, False)
        # models['analysis'].updateStateField()
        # Eout = models['analysis'].getEnsemble(Eout, N, para)
        Y[:] = obs(Eout, models['analysis'])

        para.comm_world.Bcast(Y, root=para.rank_world_filterpe)
        if para.PE['filter']:
            Eout = np.asfortranarray(Eout)
        para.comm_world.Scatterv([Eout, chunks*N], E0, root=para.rank_world_filterpe)

        _e[0] = 0.
        for l_domain in range(chunk):
            global_domain = l_domain + offset
            l_obs = tapering[global_domain][0]
            coeffs = tapering[global_domain][1]

            ym = Y[:, l_obs].sum(axis=0)/N
            Ysqrt = (Y[:, l_obs] - ym)*coeffs/eps/sig_obs[l_obs]
            delta = (y[l_obs] - ym)*coeffs/sig_obs[l_obs]

            # Compute the gradient
            gradient = (N-1.)*w0[:, l_domain] - dot(Ysqrt, delta)

            # Approximate the Hessian
            H = (N-1)*identity(N) + dot(Ysqrt, Ysqrt.T)

            # Compute the increment
            dw0 = np.linalg.solve(H, gradient)
            w0[:, l_domain] -= dw0

            _e[0] += sqrt(dot(dw0, dw0)/N)

        para.comm_world.Allreduce(_e, _e_total, op=MPI.SUM)
        _e_total = _e_total/M
        # _e = para.comm_world.bcast(_e, root=para.rank_world_filterpe)
        if p == 39:
            print("  ", para.rank_world, l_domain, chunk, p, _e_total)
        if _e_total[0] < e:
            print("  ", para.rank_world, l_domain, chunk, p, _e_total)
            break

    for l_domain in range(chunk):
        global_domain = l_domain + offset
        l_obs = tapering[global_domain][0]
        coeffs = tapering[global_domain][1]

        ym = Y[:, l_obs].sum(axis=0)/N
        Ysqrt = (Y[:, l_obs] - ym)*coeffs/eps/sig_obs[l_obs]

        x0[l_domain] = xm0[l_domain] + dot(w0[:, l_domain], X0[:, l_domain])
        H = (N-1.)*identity(N) + dot(Ysqrt, Ysqrt.T)

        U, s, V = np.linalg.svd(H, full_matrices=False)
        T = dot(U, dot(diag(sqrt((N-1.)/s)), U.T))
        E[:, l_domain] = x0[l_domain] + dot(T, X0[:, l_domain])
        _x0[l_domain] = x0[l_domain]

    # gather E
    para.comm_world.Gatherv(E, [Eout, chunks*N], root=para.rank_world_filterpe)
    if para.PE['filter']:
        Eout = np.ascontiguousarray(Eout)
    # forecast.State(model_handlers['forecast'], _x0, dim, u, v, t, Lc)
    # model.OutputGen(model_handlers['forecast'], _x0, dim, t+1)

    return 1., x0, Eout, _x0


# ##############################################################
# # SDA, S=1, Bundle, IEnKS-N                                  #
# ##############################################################
def Analysis2(int t,
              np.ndarray[np.float64_t, ndim=1] y,
              E_in,
              obs,
              tapering,
              int N, int M,
              models, para, u,
              np.ndarray[np.float64_t, ndim=1] sig_obs, int Lc):

    cdef np.ndarray[np.float64_t, ndim=2] E, Eout, Y, E0
    cdef np.ndarray[np.int_t, ndim=1] chunks
    cdef int chunk, remainder, nPE, rank, i, D, offset

    rank = para.rank_world
    nPE = para.size_world
    chunk = M//nPE
    chunks = np.ones(nPE, dtype=np.int)*chunk
    remainder = M - chunk*nPE
    chunks[:remainder] = chunks[:remainder] + 1
    if rank < remainder:
        chunk += 1
    offset = np.sum(chunks[:rank])


    D = len(sig_obs)
    if para.PE['filter']:
        Eout = np.asfortranarray(E_in)
    else:
        Eout = None

    E = np.empty((N, chunk), dtype=np.float64, order='F')
    E0 = np.empty((N, chunk), dtype=np.float64, order='F')
    Y = np.empty((N, D), dtype=np.float64)
    para.comm_world.Scatterv([Eout, chunks*N], E0, root=para.rank_world_filterpe)

    cdef double Ne = N+1.
    cdef double eN = 1. + 1./N

    # Determine size of the ensemble
    cdef np.ndarray[np.float64_t, ndim=2] w0 = zeros((N, chunk))
    cdef np.ndarray[np.float64_t, ndim=1] xm0 = zeros(chunk)
    cdef np.ndarray[np.float64_t, ndim=2] X0 = np.zeros((N, chunk))
    cdef np.ndarray[np.float64_t, ndim=1] _x0 = np.zeros_like(xm0)
    cdef np.ndarray[np.float64_t, ndim=1] x0 = np.zeros_like(xm0)

    xm0 = E0.sum(axis=0)/N
    X0 = E0 - xm0

    cdef int l_domain, p, global_domain
    cdef np.ndarray[np.float64_t, ndim=1] ym, delta, gradient, coeffs, dw0, s
    cdef np.ndarray[np.float64_t, ndim=2] Ysqrt, H, U, V, T
    cdef np.ndarray[np.int_t, ndim=1] l_obs
    cdef np.ndarray[np.float64_t, ndim=1] _e, _e_total
    _e = np.zeros(1, dtype=np.float64) 
    _e_total = np.zeros(1, dtype=np.float64)

    cdef double zeta

    if rank == nPE - 1:
        tol = np.zeros(p_max)
        tol_Ca = np.zeros(p_max)
        
        Ca = np.zeros((p_max, N))
    if rank == 0:
        tol_phys = np.zeros(p_max)

    for p in range(p_max):
        for l_domain in range(chunk):
            x0[l_domain] = xm0[l_domain] + dot(w0[:, l_domain], X0[:, l_domain])
            # Compute the sensitivities
            E0[:, l_domain] = x0[l_domain] + eps*X0[:, l_domain]

        # gather E
        para.comm_world.Gatherv(E0, [Eout, chunks*N], root=para.rank_world_filterpe)
        if para.PE['filter']:
            Eout = np.ascontiguousarray(Eout)

        models['analysis'].assignEnsemble(Eout, t, para)
        models['analysis'].scheme(u, Lc, False)
        models['analysis'].updateStateField()
        Eout = models['analysis'].getEnsemble(Eout, N, para)
        Y[:] = obs(Eout, models['analysis'])

        para.comm_world.Bcast(Y, root=para.rank_world_filterpe)
        if para.PE['filter']:
            Eout = np.asfortranarray(Eout)
        para.comm_world.Scatterv([Eout, chunks*N], E0, root=para.rank_world_filterpe)

        _e[0] = 0.
        for l_domain in range(chunk):
            global_domain = l_domain + offset
            l_obs = tapering[global_domain][0]
            coeffs = tapering[global_domain][1]

            ym = Y[:, l_obs].sum(axis=0)/N
            Ysqrt = (Y[:, l_obs] - ym)*coeffs/eps/sig_obs[l_obs]
            delta = (y[l_obs] - ym)*coeffs/sig_obs[l_obs]

            # Compute the gradient
            zeta = Ne/(eN + dot(w0[:, l_domain], w0[:, l_domain]))
            gradient = zeta*w0[:, l_domain] - dot(Ysqrt, delta)

            # Approximate the Hessian
            H = (Ne-1)*identity(N) + dot(Ysqrt, Ysqrt.T)

            # Compute the increment
            dw0 = np.linalg.solve(H, gradient)
            w0[:, l_domain] -= dw0

            _e[0] += sqrt(dot(dw0, dw0)/N)

        para.comm_world.Allreduce(_e, _e_total, op=MPI.SUM)
        _e_total = _e_total/M
        if rank == nPE - 1:
            tol_Ca[p] = sqrt(dot(dw0, dw0)/N)
            tol[p] = _e_total
            Ca[p] = xm0[-1] + dot(w0[:, -1], X0[:, -1]) + eps*X0[:, -1]
        if rank == 0:
            tol_phys[p] = sqrt(dot(dw0, dw0)/N)

        if p == 39:
            print("  ", para.rank_world, l_domain, chunk, p, _e_total)
        if _e_total[0] < e:
            print("  ", para.rank_world, l_domain, chunk, p, _e_total)
            break

    if rank == 0:
        np.savez(f'tol_phys_{t}.npz', tol_phys=tol_phys)
    if rank == nPE-1:
        np.savez(f'tol_{t}.npz', tol=tol, Ca=Ca, tol_Ca=tol_Ca)

    for l_domain in range(chunk):
        global_domain = l_domain + offset
        l_obs = tapering[global_domain][0]
        coeffs = tapering[global_domain][1]

        ym = Y[:, l_obs].sum(axis=0)/N
        Ysqrt = (Y[:, l_obs] - ym)*coeffs/eps/sig_obs[l_obs]

        x0[l_domain] = xm0[l_domain] + dot(w0[:, l_domain], X0[:, l_domain])
        zeta = Ne/(eN + dot(w0[:, l_domain], w0[:, l_domain]))
        H = zeta*identity(N) - 2*zeta*zeta/Ne*dot(w0[:, l_domain][newaxis].T, w0[:, l_domain][newaxis]) \
            + dot(Ysqrt, Ysqrt.T)

        U, s, V = np.linalg.svd(H, full_matrices=False)
        T = dot(U, dot(diag(sqrt((N-1.)/s)), U.T))
        E[:, l_domain] = x0[l_domain] + dot(T, X0[:, l_domain])
        _x0[l_domain] = x0[l_domain]

    # gather E
    para.comm_world.Gatherv(E, [Eout, chunks*N], root=para.rank_world_filterpe)
    if para.PE['filter']:
        Eout = np.ascontiguousarray(Eout)

    # forecast.State(model_handlers['forecast'], _x0, dim, t, Lc)

    return zeta, x0, Eout, _x0


##############################################################
#  Inflation of the ensemble                                 #
##############################################################
def Inflation(E, infl):
    # Determine size of the ensemble
    N = E.shape[0]

    # Compute the mean
    xm = E.sum(axis=0)/N

    # Rescale
    E[:N] = xm + infl*(E - xm)

