#!/usr/bin/env python
# -*- coding: iso-8859-15 -*-h


############################################################
#
#      (c)   Marc Bocquet, ENPC/CEREA
#      Programme ienks_call.py
#      Initial release: 22 June 2015
#      Latest modification: 22 June 2015
#
############################################################

import time
from collections import deque
import numpy as np
import functools
from mpi4py import MPI

import MEB_online as model
from Observation import Observation
import ienks_analysis
import local


def ienks_call(para, Nt, N, L, S, infl, dl_radius, gridTypes, doN=False, IAU=False, nIAU=0):

    ########################
    #  Setup and Spinup    #
    ########################

    # initialise models

    # we need three different models:
    # 1) ensemble 2) inner loop model ('analysis') 3) forecast
    models = {key: model.MEB(para, key) for key in ('ensemble', 'analysis', 'forecast')}
    # temporal information
    Ntsu = models['ensemble'].nt
    dt = 30

    # initialise wind forcing
    storm_ens, storm_t = models['ensemble'].initWind(para.task_id, True, True)

    # get the size of the state vector and the observations
    M = models['ensemble'].StateSize
    P = models['ensemble'].ParamSize

    E = None
    if para.PE['filter']:
        E = np.empty([N, M], dtype='d')

    # initialise observations
    obs_ens = Observation(models['ensemble'], gridTypes)
    D = obs_ens.ObsDim

    tapering = [local.getTaper(l_domain, obs_ens, models['ensemble'], dl_radius, cutoff=1e-5) for l_domain in range(M - P)]
    tapering += [(np.arange(obs_ens.ObsDim), np.ones(obs_ens.ObsDim)) for i in range(P)]

    # Current length of the data assimilation window (Lc<=L)
    Lc = 0
    # Current shift of the data assimilation window (Sc<=S)
    Sc = S
    # Lag between the observation and the analysis time (actual DA window)
    offset = 0

    # initial wind forcing
    assert para.PE['ensemble'], 'storm ensemble is on the wrong PEs'
    u = deque([storm_ens.updateWind(dt, it + Ntsu + offset + 1, dt) for it in range(Lc)])
    u_t = deque([storm_t.updateWind(dt, it + Ntsu + offset + 1, dt) for it in range(Lc)])

    iIAU = np.inf
    # ###########################
    # #  Data Assimilation Run  #
    # ###########################
    for t in range(Nt):

        if t % Sc == 0:
            # get ensemble
            para.comm_world.Barrier()

            models['ensemble'].updateField()
            E = models['ensemble'].getEnsemble(E, N, para)

            if models['ensemble'].X is not None:
                models['analysis'].X = models['ensemble'].X

            if not doN:
                if para.PE['filter']:
                    # Compute the mean
                    xm = E[:, :-1].sum(axis=0)/N
                    # Rescale
                    E[:, :-1] = xm + infl*(E[:, :-1] - xm)

                    i_infl = -len(model.global_params_state)
                    for param_name in model.global_params_state:
                        xm = E[:, i_infl].sum(axis=0)/N
                        infl_Ca= max(0.4*model._params_std[param_name]/np.std(E[:, i_infl], axis=0), 1.)
                        print ('infl_Ca:', infl_Ca)
                        E[:, i_infl] = xm + infl_Ca*(E[:, i_infl] - xm)
                        i_infl += 1


            # get observations
            y = np.empty(D, dtype=np.float64)
            sig_obs = np.empty(D, dtype=np.float64)
            y, sig_obs = obs_ens.genSynthetic(para, t + Ntsu + Lc + offset, y, sig_obs)
            models['ensemble'].SaveParams(E, t + Ntsu + offset, para, 'for')

            if para.PE['filter'] and IAU:
                E0 = E.copy()

            # analysis for the state
            if doN:
                zeta, x0, E, _x0 = ienks_analysis.Analysis2(t + Ntsu + offset, y, E, obs_ens.H, tapering, N, M, models, para, u, sig_obs, Lc)
            else:
                zeta, x0, E, _ = ienks_analysis.Analysis1(t + Ntsu + offset, y, E, obs_ens.H, tapering, N, M, models, para, u, sig_obs, Lc)
            
 
            # analysis for the parameter
            models['analysis'].OutputGenE(E, t + Ntsu + offset, para)
            if para.PE['filter'] and IAU:
                increment = E - E0
            iIAU = 0

            if not IAU:
                models['ensemble'].mean_params = models['analysis'].mean_params
                models['ensemble'].assignEnsemble(E, t + Ntsu + offset, para)

            if para.PE['filter']:
                print(" - step", t, "Lc", Lc, "Sc", Sc, "%3.2f" % zeta)

            if models['ensemble'].dg is not None:
                models['ensemble'].dg.change_iteration(100)

            models['ensemble'].SaveParams(E, t + Ntsu + offset, para, 'ana')


        if para.PE['ensemble']:
            u.append(storm_ens.updateWind(dt, t + Lc + Ntsu + offset + 1, dt))
            u_t.append(storm_t.updateWind(dt, t + Lc + Ntsu + offset + 1, dt))

        if Lc == L:
            if IAU and iIAU < nIAU:
                E = models['ensemble'].getEnsemble(E, N, para)
                if para.PE['filter']:
                    E = E + increment/nIAU
                models['ensemble'].assignEnsemble(E, t + Ntsu + offset, para)
                iIAU += 1

            # propagate the ensemble
            models['ensemble'].scheme(u, u_t, 1, para.comm_couple, True)

            if (t + 120) % Sc == 0:
                if models['ensemble'].dg is not None:
                    models['ensemble'].dg.change_iteration(100)
            if (t - 120) % Sc == 0:
                if models['ensemble'].dg is not None:
                    models['ensemble'].dg.change_iteration(10)

            if para.PE['ensemble']:
                u.popleft()
                u_t.popleft()
        else:
            Lc += 1
            offset -= 1
            if Lc == L:
                Sc = S
