"""Summary
"""
import numpy as np
import pandas as pd

_day_in_seconds = 24*3600
_length = 1e3


class Storm:
    """Generating wind storm for each time step
    
    Attributes
    ----------
    a : float
        strength coefficient
    b : float
        strength power
    dx : float
        random walk distance
    L : float
        Length of the domain
    storm_period : iterator
        An iterator of T for the simulation time
    T : dict
        a dictionary of the final time step of the storm,
        storm period, and if the storm exists
    Ud : float
        Maximum divergent wind speed
    Us : float
        Maximum rotational wind speed
    vb : float
        background wind speed
    vc : float
        velocity of the storm in the reference coordinate
    x : ndarray
        reference coordinate
    xc : float
        center of the storm
    y : ndarray
        reference coordinate
    yc : float
        center of the storm
    
    """

    def __init__(self, task_id, x, y, Lmin, L, dt, nt, do_random, do_walk):
        """Storm constructor
        
        Parameters
        ----------
        task_id : int
            ensemble id
        x : ndarray
            an array of grid points
        y : ndarray
            an array of grid points
        Lmin : float
            starting position of model domain
        L : float
            length of model domain
        dt : float
            time resolution
        nt : float
            total simulation time in seconds
        """
        self.do_random = do_random
        self.do_walk = do_walk
        if not self.do_random:
            do_walk = False
        self.task_id = task_id
        # get reference coordinate
        self.x_in = x
        self.y_in = y
        self.L = L
        self.Lmin = Lmin
        self.initNewStorm(0)

        # time tapering parameter
        self.a = 10
        self.b = 2

        # storm period
        T = self.getT(task_id, nt)
        it = (np.cumsum(T)/dt).astype(int)
        is_storm = np.ones(len(T), dtype=int)
        is_storm[1::2] = 0
        self.storm_period = pd.DataFrame(zip(it, T, is_storm),
                                         columns=['timestep',
                                                  'period',
                                                  'is_storm'],
                                         dtype='object'
                                         ).astype({'timestep': 'int',
                                                   'period': 'float',
                                                   'is_storm': 'int'
                                                   }
                                                  ).to_dict('records')
        self.storm_period = iter(self.storm_period)
        self.T = next(self.storm_period)
        self.countStorm = 0


    def initNewStorm(self, iStorm):
        # get reference coordinate
        self.xc, self.yc = self.getCenter(self.task_id, iStorm, self.Lmin, self.L)
        self.x, self.y = self.ref_coord(self.xc, self.yc,
                                        self.Lmin, self.L, self.x_in, self.y_in)
        self.xc = (self.xc - self.Lmin)/self.L
        self.yc = (self.yc - self.Lmin)/self.L


        # storm speed
        self.vc = self.getVelocity(self.task_id, iStorm, self.L)
        self.dx = None
        if self.do_walk:
            self.x = self.randomWalk(iStorm, 0., self.task_id, self.x, self.L)
        self.Ud = 0.1
        # get Us
        if self.do_random:
            np.random.seed((iStorm*100 + self.task_id) % 300000)
            self.Us = np.random.normal(22, 0.5)
            # background velocity
            np.random.seed((iStorm*100 + self.task_id) % 300000)
            self.vb = np.random.normal(2, np.sqrt(0.05))
        else:
            self.Us = 22.
            self.vb = 2.

    def updateWind(self, dt, it, dt_walk):
        """Compute wind at time t
        
        Parameters
        ----------
        dt : float
            Random walk for dx time interval
        it : int
            current time step
        dt_walk : float
            time interval of random walk
        task_id : int
            ensemble id
        
        Returns
        -------
        ndarray
            Wind field in both x and y direction
        """
        t = it*dt

        if it > self.T['timestep']:
            self.T = next(self.storm_period)
            self.countStorm += 1
            self.initNewStorm(self.countStorm)

        if self.T['is_storm']:
            self.y = self.storm_travel(self.y, self.vc, dt)
            if self.do_walk:
                self.x = self.randomWalk(it, dt_walk, self.task_id, self.x, self.L)

            us, vs = self.u_r(self.Us, self.x, self.y)
            ud, vd = self.u_d(self.Ud, self.x, self.y)

            mask = self.mask(self.x, self.y)

            # print(f'...Current storm period for task {task_id} period is: '
            #       f'{self.T["period"]/_day_in_seconds} days...')

            gamma = self.gamma(self.a, self.b, self.T['period'], t - (self.T['timestep']*dt - self.T['period']))
            return (us + ud)*gamma*mask, \
                   (vs + vd)*gamma*mask + self.vb
        else:
            return np.zeros(self.x.shape), np.zeros(self.y.shape) + self.vb

    def getT(self, task_id, nt):
        """generating period of storms
        
        Parameters
        ----------
        task_id : int
            current ensemble id of the member
        nt : int
            total days of experiment
        
        Returns
        -------
        ndarray
            an array of time period in the experiment
        """
        # a list of storm period
        np.random.seed(0)
        ndays = nt//_day_in_seconds
        T = np.random.uniform(2, 5, size=365)

        if self.do_random:
            np.random.seed(task_id)
            dT = np.random.normal(0, 0.5, size=365)
        else:
            dT = 0.

        T = T + dT

        if ndays >= 2:
            index = np.cumsum(T) >= ndays
            index = np.min(np.where(index)[0])
            T = T[:index+1]
        else:
            T = np.array([T[0]])

        return T*_day_in_seconds

    def getCenter(self, task_id, iStorm, Lmin, L):
        """Position of the eye of the storm
        
        Parameters
        ----------
        task_id : int
            ensemble id
        Lmin : float
            starting position of model domain
        L : float
            length of model domain
        
        Returns
        -------
        ndarray
            ndarray of storm center at initial time
        """
        np.random.seed(iStorm % 300000)
        yc = np.random.uniform(-50, 20)
        if self.do_random:
            np.random.seed((task_id*100 + iStorm) % 300000)
            dyc = np.random.normal(0, 1.5)
        else:
            dyc = 0.
        yc = (yc + dyc)*_length


        np.random.seed(iStorm % 300000)
        xc_base = np.random.uniform(-25, 25)
        if self.do_random:
            np.random.seed((task_id*200 + iStorm) % 300000)
            dxc = np.random.normal(0, 1.5)
        else:
            dxc = 0.
        xc = (xc_base + dxc)*_length
        i = 1
        d = (xc - Lmin)/L
        while d < 0.3 or 1 - d < 0.25:
            assert self.do_random, 'storm center should be within limit without randomness'
            np.random.seed((task_id*20 + iStorm + 100*i) % 300000)
            xc = (xc_base + np.random.normal(0, 1.5))*_length
            d = (xc - Lmin)/L
            i += 1
        return xc, yc

    def getVelocity(self, task_id, iStorm, L):
        """velocity of the storm movement
        
        Parameters
        ----------
        task_id : int
            ensemble id
        L : float
            length of the domain
        
        Returns
        -------
        float
            velocity of the storm in the reference coordinate
        """
        vc = 0.25
        if self.do_random:
            np.random.seed((iStorm + task_id*100) % 300000)
            vc += np.random.normal(0, 0.0045)
        return vc/L

    def randomWalk(self, it, dt, task_id, x, L):
        """update zonal random walk
        
        Parameters
        ----------
        t : float
            current time
        dt : float
            time interval for random walk
        task_id : int
            ensemble id
        x : ndarray
            gridpoints on reference coordinate [0, 1]
        L : float
            Domain length
        
        Returns
        -------
        ndarray
            Random walked coordinates
        """
        if self.dx is None:
            np.random.seed( (int(it) + task_id*30) % 300000)
            self.dx = np.random.normal(0, 0.01/225.)
            self.xc += self.dx
            return x + self.dx

        np.random.seed((int(it) + task_id*100) % 300000)
        tau = 60.
        dx = np.exp(-dt/tau)*self.dx + np.random.normal(0, 0.01/225)
        i = 1
        while self.xc + dx < 0.3 or 1 - (self.xc + dx) < 0.25:
            np.random.seed((task_id + 1000 + 100*i) % 300000)
            dx = np.exp(-dt/tau)*self.dx + np.random.normal(0, 0.01/225)
            i += 1
        self.dx = dx
        self.xc += self.dx
        return x + self.dx

    def u_r(self, Us, x, y):
        """rotational wind
        
        Parameters
        ----------
        Us : float
            Maximum rotational wind speed
        x : ndarray
            gridpoints on reference coordinate [0, 1]
        y : ndarray
            gridpoints on reference coordinate [0, 1]
        
        Returns
        -------
        ndarray
            rotational wind field
        """
        u = Us*np.power(np.sin(2*np.pi*x), 2)*np.sin(4*np.pi*y)
        v = -Us*np.power(np.sin(2*np.pi*y), 2)*np.sin(4*np.pi*x)
        return u, v

    def u_d(self, Ud, x, y):
        """divergent wind
        
        Parameters
        ----------
        Ud : float
            Maximum divergent wind speed
        x : ndarray
            gridpoints on reference coordinate [0, 1]
        y : ndarray
            gridpoints on reference coordinate [0, 1]
        
        Returns
        -------
        ndarray
            divergent wind field
        """
        u = Ud*np.power(np.sin(2*np.pi*y), 2)*np.sin(4*np.pi*x)
        v = Ud*np.power(np.sin(2*np.pi*x), 2)*np.sin(4*np.pi*y)
        return u, v

    def gamma(self, a, b, T, t):
        """Attenuation factor for storm
        
        Parameters
        ----------
        a : float
            strength coefficient
        b : float
            strength power
        T : float
            storm period
        t : float
            current time
        
        Returns
        -------
        float
            storm strength attenuation factor
        """
        strength = np.power(np.abs(np.sin(np.pi*t/T)), b)
        return 1 - np.exp(-a*strength)

    def storm_travel(self, y, v, dt):
        """Relative displacement of the domain
        
        Parameters
        ----------
        y : ndarray
            an array of grid point positions
        v : float
            velocity of the storm traversal in y-direction
        dt : TYPE
            Description
        
        Returns
        -------
        ndarray, ndarray
            new position of the domain
        
        Deleted Parameters
        ------------------
        u : float
            velocity of the storm traversal in x-direction
        t : float
            current time
        """
        y = y - v*dt
        return y

    def ref_coord(self, xc, yc, Lmin, L, x, y):
        """convert to reference coordinate [0, 1]
        
        Parameters
        ----------
        xc : float
            center of the storm
        yc : float
            center of the storm
        Lmin : float
            starting position of model domain
        L : float
            length of model domain
        x : ndarray
            an array of grid points
        y : ndarray
            an array of grid points
        
        Returns
        -------
        ndarray
            reference coordinates
        """
        x = ((x - xc) - Lmin)/L - 0.25
        y = ((y - yc) - Lmin)/L - 0.25
        return x, y

    def mask(self, x, y):
        """Mask out additional storms/waves
        
        Parameters
        ----------
        x : ndarray
            gridpoints on reference coordinate [0, 1]
        y : ndarray
            gridpoints on reference coordinate [0, 1]
        
        Returns
        -------
        ndarray
            ndarray of bool
        """
        mask_u = (1 - np.fmod(np.floor(x), 2))*(np.sin(2*np.pi*x) > 0)
        mask_v = np.sin(2*np.pi*y) > 0
        return mask_u*mask_v


if __name__ == '__main__':
    dt = 30
    x = np.arange(-100000, 100000, 8000)
    y = np.arange(-100000, 100000, 8000)
    X, Y = np.meshgrid(x, y)
    Lmin = -100000
    L = 200000
    nt = 90

    from mpi4py import MPI
    import matplotlib.pyplot as plt
    import time

    task_id = MPI.COMM_WORLD.Get_rank() + 1
    storm = Storm(task_id, X, Y, Lmin, L, dt, 30*24*3600, True, True)
    fig = plt.figure(str(task_id))
    ax = fig.add_subplot(111)
    start = time.time()
    gammas = []
    for i in range(24*3600):
        u, v, gamma = storm.updateWind(dt, i, dt)
        gammas.append(gamma)
    plt.plot(np.arange(len(gammas)), np.asarray(gammas))
    plt.show()
    #     if i%225 == 0:
    #         plt.ion()
    #         ax.cla()
    #         norm = np.sqrt(u*u + v*v)
    #         ax.quiver(X, Y, u*norm, v*norm)
    #         ax.set_title(str(i*dt/_day_in_seconds))
    #         plt.pause(0.001)
    # end = time.time()
    # print(end - start)
    # plt.show()


