"""Example script of functions in pyMEB interface
"""
import numpy as np
from mpi4py import MPI

import pyMEB
import wind
import pathlib
import os
import glob
import shutil

# parallelisation set-ups
n_model_PE = 1
comm_world = MPI.COMM_WORLD
rank_world = comm_world.Get_rank()
size_world = comm_world.Get_size()

# split the CPUs for each ensemble
color = rank_world//n_model_PE
task_id = color + 1
comm_model = comm_world.Split(color, rank_world)
rank_model = comm_model.Get_rank()
size_model = comm_model.Get_size()

# basic dimensions
total_steps = 90*24*3600
Lmin = - 100000
L = 200000
# reset model n_loop times to avoid memory overflow
n_loops = total_steps//24//3600//3
dt = 30
# time steps for each loop
nt = total_steps//dt//n_loops
print(n_loops, nt)

Cfix = 5000
output_dir = f'output{task_id}_{Cfix}'
input_dir = f'input{task_id}'
home = str(pathlib.Path.home())
# model options
argv = ['./run_pyMEB',
        f'--config={home}/MEB_DG/model-config/EB_channel_adv.cfg',
        f'--parameter.C_fix={Cfix}',
        f'--resolution.nt={nt}',
        f'--input.dirname={input_dir}',
        f'--output.dirname={output_dir}',
        f'--input.restartfile-timeindex={nt//120}',
        '--forcing.forcing-type=External']

# init model
dg, post, output, dt, d = pyMEB.initModel(comm_model, argv)
# get mesh coord
x, y = pyMEB.getVarCoordinate(dg, 'forcing_x')
# initialise storm
storm = wind.Storm(task_id, x, y, Lmin, L, dt, total_steps, True, True)
# selecting initial time step based on the index of the output file
# for example, output_1 comes from the first n_loops.
# output_2 comes from the 2-nd n_loops
files = glob.glob(f'{output_dir}/output_*.nc')
if len(files) == 0:
    start = 0
else:    
    start = np.max([int(f.split('_')[-1].split('.')[0]) for f in files])
    argv.append('--input.from-restart=true')

if rank_model == 0:
    print(f'task {task_id} will start from the {start}-th loop')

# update wind forcing state to starting point
for i_times in range(start):
    for i in range(nt):
        storm.updateWind(dt, nt*i_times + i + 1, dt)


# starting model forward
for i_times in range(start, n_loops):
    # ensure the model using restart file
    if i_times == 1 and start != 1:
        argv.append('--input.from-restart=true')
    # delete objects to reset model
    del dg, post, output, dt, d
    comm_model.barrier()
    # remove output.nc and forcing.nc to prevent errors
    if i_times == start and rank_model == 0:
        os.remove(f'{output_dir}/output.nc')
        os.remove(f'{output_dir}/forcing.nc')
    comm_model.barrier()
    # initialise the model
    dg, post, output, dt, d = pyMEB.initModel(comm_model, argv)
    dg.set_taskid(task_id)
    if i_times == 0:
        dim = pyMEB.getArraySize(dg, 'u')
        var = np.zeros(dim)
        pyMEB.getValue(dg, 'u', var)
        var += np.random.normal(loc=0.0, scale=0.05, size=dim)
        pyMEB.assignValue(dg, 'u', var)

        dim = pyMEB.getArraySize(dg, 'v')
        var = np.zeros(dim)
        pyMEB.getValue(dg, 'v', var)
        var += np.random.normal(loc=0.0, scale=0.05, size=dim)
        pyMEB.assignValue(dg, 'v', var)

        x, y = pyMEB.getVarCoordinate(dg, 'thickness')
        dim = pyMEB.getArraySize(dg, 'thickness')
        var = np.zeros(dim)
        r0 = 50000
        r = np.sqrt(x*x + y*y)
        var = np.where(r<r0, 1+np.cos(0.5*np.pi*r/r0), 1.)
        pyMEB.assignValue(dg, 'thickness', var)

    Ca = np.load(f'{output_dir}/Ca.npy')[task_id - 1]
    pyMEB.assignValueSerial(dg, 'C_a', Ca)
    pyMEB.adjustModelConsistency(dg, nt*i_times)

    for i in range(nt):
        if rank_model == 0:
            print('Current time step:', i, 'for task:', task_id)
        # setting wind field
        u, v = storm.updateWind(dt, nt*i_times + i + 1, dt)
        pyMEB.assignValue(dg, 'forcing_x', u)
        pyMEB.assignValue(dg, 'forcing_y', v)

        # forward model
        pyMEB.stepModel(dg, post, output, d)
        output.update(dg, post)
    comm_model.barrier()

    # rename output and forcing files based on the loop index
    if rank_model == 0:
        os.rename(f'{output_dir}/output.nc', f'{output_dir}/output_{i_times+1}.nc')
        os.rename(f'{output_dir}/forcing.nc', f'{output_dir}/forcing_{i_times+1}.nc')
        shutil.copy(f'{output_dir}/output_{i_times+1}.nc', f'{input_dir}/restart.nc')

del dg, post, output, dt, d
comm_model.barrier()

